/*
 * BigDouble.cpp
 *
 *  Created on: Mar 1, 2012
 *      Author: jeffrp
 */

#include "BigDouble.h"


BigDouble::BigDouble() {
	this->exponent = 0;
	this->mantissa = 0.0l;
}


BigDouble::BigDouble(const long double value) {
	this->exponent = 0;
	this->mantissa = value;
	this->adjustMantissa();
}

BigDouble::BigDouble(const long double mantissa, const int exponent) {
	this->exponent = exponent;
	this->mantissa = mantissa;
}

BigDouble::~BigDouble() {

}

//BigDouble::operator long double() {
//	return ldexp(this->mantissa, this->exponent);
//}





BigDouble BigDouble::operator -(const BigDouble& operand)
{
	BigDouble negOperand(-operand.mantissa, operand.exponent);
	return *this + negOperand;
}

void BigDouble::operator -=(const BigDouble& operand) {
	BigDouble result = *this - operand;
	this->exponent = result.exponent;
	this->mantissa = result.mantissa;
}

BigDouble BigDouble::operator *(const BigDouble& operand)
{
	BigDouble result(this->mantissa, this->exponent);
	result.mantissa *= operand.mantissa;
	result.exponent += operand.exponent;
	result.adjustMantissa();
	return result;
}

BigDouble BigDouble::operator *(const long double operand)
{
	BigDouble result(this->mantissa, this->exponent);
	result.mantissa *= operand;
	result.adjustMantissa();
	return result;
}

void BigDouble::operator *=(const BigDouble& operand)
{
	BigDouble result = *this * operand;
	this->exponent = result.exponent;
	this->mantissa = result.mantissa;
}


BigDouble BigDouble::operator +(const BigDouble& operand)
{
	BigDouble result;
	int expDiff = this->exponent - operand.exponent;
	if ((expDiff > 60  && this->mantissa != 0.0) || operand.mantissa == 0.0) {
		result.exponent = this->exponent;
		result.mantissa = this->mantissa;
	}
	else if (expDiff < -60 || this->mantissa == 0.0) {
		result.exponent = operand.exponent;
		result.mantissa = operand.mantissa;
	}
	else {
		result.mantissa = this->mantissa + ldexp(operand.mantissa, -expDiff);	// operand.mantissa * (2 ^ expDiff);
		result.exponent = this->exponent;

		result.adjustMantissa();
	}

	return result;
}

void BigDouble::operator +=(const BigDouble& operand) {
	BigDouble result = *this + operand;
	this->exponent = result.exponent;
	this->mantissa = result.mantissa;
}

void BigDouble::adjustMantissa() {
	int newExponent;
	this->mantissa = frexp(this->mantissa, &newExponent);
	this->exponent += newExponent;
}


void BigDouble::operator =(const long double operand) {
	this->mantissa = operand;
	this->exponent = 0;
	this->adjustMantissa();
}


void BigDouble::operator =(const BigDouble& operand) {
	this->mantissa = operand.mantissa;
	this->exponent = operand.exponent;
}

std::ostream& operator << (std::ostream& out_file, const BigDouble& bd) {
	out_file << ldexp(bd.mantissa, bd.exponent);
	return(out_file);
}


long double BigDouble::logarithm(long double base) {
	return (log(this->mantissa) + log(2.0l) * this->exponent)/log(base);
}

double log(BigDouble number) {
	if (number.mantissa == 0.0) {
		cerr << "BigDouble warning: calculating logarithm of zero" << endl;
	}
	return (log(number.mantissa) + log(2.0l) * number.exponent);
}

double log10(BigDouble number) {
	return ((double) number.logarithm(10.0));
}



int test() {

	//BigDouble bd1(0.003423l);
	//BigDouble bd1;
	//bd1 = 0.003423;


	//BigDouble bd2(0.125123l);

	/*
	//cout << (bd1 + bd2) << endl;
	//cout << (bd1 * bd2) << endl;
	cout << (bd1 - bd2) << endl;
	cout << bd1.logarithm(10.0l) << endl;
	cout << log10(bd1) << endl;
	cout << log(bd1)/log(2) << endl;
	cout << (bd1 + 0.07412l) << endl;
	cout << (bd1 + 0.07412) << endl;
	cout << (bd1 * 0.07412l) << endl;
	cout << (bd1 * 0.07412) << endl;
	//cout << (0.07412l * bd1) << endl;
	//cout << (0.07412 * bd1) << endl;
	*/

	double val1, val2;
	BigDouble bd1, bd2;

	val1 = 0.00047629831879;
	val2 = 0.00000008732978;
	bd1 = val1;
	bd2 = val2;
	cout << "normal addition: " << (val1 + val2) << endl;
	cout << "BD addition: " << (bd1 + bd2) << endl;
	cout << "normal multiplication " << (val1 * val2) << endl;
	cout << "BD multiplication " << (bd1 * bd2) << endl;
	bd1 += bd2;
	cout << "BD shortcut addition-assignment " << bd1 << endl;
	cout << "normal log " << log(val2) << endl;
	cout << "BD log " << log(bd2) << endl;
	cout << endl;

	val1 = 0.000000000000000000000000003965293847926293791828;
	val2 = 0.00000000000000000000000000000000000000234876234;
	bd1 = val1;
	bd2 = val2;
	cout << "normal addition: " << (val1 + val2) << endl;
	cout << "BD addition: " << (bd1 + bd2) << endl;
	cout << "normal multiplication " << (val1 * val2) << endl;
	cout << "BD multiplication " << (bd1 * bd2) << endl;
	bd1 += bd2;
	cout << "BD shortcut addition-assignment " << bd1 << endl;
	cout << "normal log " << log(val2) << endl;
	cout << "BD log " << log(bd2) << endl;
	cout << endl;

	val1 = 0.00239486273549283674823764928376473829387457898;
	val2 = 0.00000000000000000000000000000000000000234876234;
	bd1 = val1;
	bd2 = val2;
	cout << "normal addition: " << (val1 + val2) << endl;
	cout << "BD addition: " << (bd1 + bd2) << endl;
	cout << "normal multiplication " << (val1 * val2) << endl;
	cout << "BD multiplication " << (bd1 * bd2) << endl;
	bd1 += bd2;
	cout << "BD shortcut addition-assignment " << bd1 << endl;
	cout << "normal log " << log(val2) << endl;
	cout << "BD log " << log(bd2) << endl;
	cout << endl;

	val1 = 0.0023948627354928367482;
	val2 = 0.0042368723487623459387;
	bd1 = val1;
	bd2 = val2;
	cout << "normal addition: " << (val1 + val2) << endl;
	cout << "BD addition: " << (bd1 + bd2) << endl;
	cout << "normal multiplication " << (val1 * val2) << endl;
	cout << "BD multiplication " << (bd1 * bd2) << endl;
	bd1 += bd2;
	cout << "BD shortcut addition-assignment " << bd1 << endl;
	cout << "normal log " << log(val2) << endl;
	cout << "BD log " << log(bd2) << endl;
	cout << endl;

	val1 = 0.0;
	val2 = 0.00000000000000000000000000000000000000234876234;
	bd1 = val1;
	bd2 = val2;
	cout << "normal addition: " << (val1 + val2) << endl;
	cout << "BD addition: " << (bd1 + bd2) << endl;
	cout << "normal multiplication " << (val1 * val2) << endl;
	cout << "BD multiplication " << (bd1 * bd2) << endl;
	bd1 += bd2;
	cout << "BD shortcut addition-assignment " << bd1 << endl;
	cout << "normal log " << log(val2) << endl;
	cout << "BD log " << log(bd2) << endl;
	cout << endl;

	bd1 = 0.0;
	cout << "normal log of zero " << log(0.0) << endl;
	cout << "BD log of zero " << log(bd1) << endl;

	return 0;

}
